/* $Header: solver.c,v 3.3 98/08/13 17:16:58 carl Exp $ */

/************************************************************************
 * Copyright (c) 1998 The Trustees of Princeton University.
 * All rights reserved.
 * 
 * PRINCETON UNIVERSITY MAKES NO REPRESENTATIONS ABOUT THE SUITABILITY OF
 * THIS SOFTWARE FOR ANY PURPOSE.  IT IS PROVIDED "AS IS" WITHOUT EXPRESS
 * OR IMPLIED WARRANTY.  PRINCETON UNIVERSITY SHALL NOT BE LIABLE FOR ANY
 * DAMAGES SUFFERED BY THE USERS OF THIS SOFTWARE.
 ************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "solver.h"
/* This is included for the function pointers for the various solvers */
#include "solver_proto.h"

static long int DEC_SOLV = 0;
/* DEClared SOLVers */
/* DEC_SOLV is a 32 bit integer containing
   information about the solvers to
   be used for various problems.
   Going RIGHT to LEFT in 8 bit portions:
   LP, NLP, MILP, MINLP, DYN */

/* Similar to the problem_id_struct in minopt_config.c */
static struct problem_type_struct {
  char *type_string; /* name of the problem type */
  int type; /* code for the problem type */
} problem_type_tab[] = {
  "LP",    LPTYPE,
  "MILP",  MILPTYPE,
  "NLP",   NLPTYPE,
  "OCP",   OCPTYPE,
  "MINLP", MINLPTYPE,
  "DYN",   DYNTYPE,
  NULL,    0
};

/* define a table to hold the information for the solvers
   for the various subproblems */
static struct solver_struct {
  char *name;/* name of the solver */
  long int code;           /* the code number for the solver (0 if not defined) */
  int cap;                 /* what the solver's capability is */
  minfuntype minfun;       /* MINLP solver */
  prifuntype prifun;       /* primal solver function */
  prifuntype iprifun;      /* infeasible primal solver */
  rlxfuntype rlxfun;       /* relaxed problem solver */
  masfuntype gbdmasfun;    /* gbd master solver */
  masfuntype oaermasfun;   /* oaer master solver */
  masfuntype oaerapmasfun; /* ap master solver */
  masfuntype dsub1fun;
  masfuntype dsub2fun;
  dmsfuntype dmsfun;
  dynfuntype dynfun;        /* dynamic solver */
  adjfuntype adjfun;        /* adjoint solver */
} solver_tab[] = {

  /* CPLEX */
  "CPLEX", CPLEXSOLV, CPLEXCAP,
  NULL,
  CPLEX_(primal_cplex), CPLEX_(iprimal_cplex), NULL,
  CPLEX_(gbd_master_cplex), CPLEX_(oaer_master_cplex), CPLEX_(oaerap_master_cplex),
  NULL, NULL, NULL,
  NULL, NULL,
  
  /* LP_SOLVE */
  "LPSOLVE", LPSOLVESOLV, LPSOLVECAP,
  NULL,
  LPSOLVE_(primal_lpsolve), LPSOLVE_(iprimal_lpsolve), NULL,
  LPSOLVE_(gbd_master_lpsolve), LPSOLVE_(oaer_master_lpsolve), LPSOLVE_(oaerap_master_lpsolve),
  NULL, NULL, NULL,
  NULL, NULL,

  /* MINOS */
  "MINOS", MINOSSOLV, MINOSCAP,
  NULL,
  MINOS_(primal_minossnopt), MINOS_(iprimal_minossnopt), MINOS_(relax_minossnopt),
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL,

  /* SNOPT */
  "SNOPT", SNOPTSOLV, SNOPTCAP,
  NULL,
  SNOPT_(primal_minossnopt), SNOPT_(iprimal_minossnopt), SNOPT_(relax_minossnopt),
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL,

  /* NPSOL */
  "NPSOL", NPSOLSOLV, NPSOLCAP,
  NULL,
  NPSOL_(primal_npsol), NPSOL_(iprimal_npsol), NPSOL_(relax_npsol),
  NPSOL_(gbd_master_npsol), NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL,

  /* FSQP */
  "FSQP", FSQPSOLV, FSQPCAP,
  NULL,
  FSQP_(primal_fsqp), FSQP_(iprimal_fsqp), NULL, 
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL,

  /*DAEOPT*/
  "DAEOPT", DAEOPTSOLV, DAEOPTCAP,
  NULL,
  DAEOPT_(primal_daeopt), DAEOPT_(iprimal_daeopt), DAEOPT_(relax_daeopt),
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL,

  /* GBD */
  "GBD", GBDSOLV, GBDCAP,
  GBD_(gbd_solve),
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL,

  /* OAER */
  "OAER", OAERSOLV, OAERCAP,
  OAER_(oaer_solve),
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL,

  /* OAERAP */
  "OAERAP", OAERAPSOLV, OAERAPCAP,
  OAERAP_(oaerap_solve),
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL,

  /* GCD */
  "GCD", GCDSOLV, GCDCAP,
  GCD_(gcd_solve),
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL,

  /* DASOLV */
  "DASOLV", DASOLVSOLV, DASOLVCAP,
  NULL,
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  DASOLV_(dynamic_dasolv), DASOLV_(adjoint_dasolv),

  /* DAESSA */
  "DAESSA", DAESSASOLV, DAESSACAP,
  NULL,
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  DAESSA_(dynamic_daessa), DAESSA_(adjoint_daessa),

  /* DASPKSO */
  "DASPKSO", DASPKSOSOLV, DASPKSOCAP,
  NULL,
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  DASPKSO_(dynamic_daspkso), DASPKSO_(adjoint_daspkso),

  /* END */
  NULL, 0, 0,
  NULL,
  NULL, NULL, NULL, 
  NULL, NULL, NULL,
  NULL, NULL, NULL,
  NULL, NULL
};

void MINOPT_print_solvers(FILE *fp)
{
  int i;
  
  fprintf(fp,"This version of MINOPT compiled with the following solvers:\n");
  for(i=0;solver_tab[i].name;i++)
    if(DOIHAVE(solver_tab[i].code))
      fprintf(fp,"\t%s\n",solver_tab[i].name);
}

void set_solver_defaults()
{
  /* Set defaults for solvers */
  SETDEFSOLV(DEFSOLV);
}

int set_solver(char *type_string, char *solver)
{
  int type,i;

  for(i=0;problem_type_tab[i].type;i++)
    if(!strcmp(problem_type_tab[i].type_string,type_string))
      type = problem_type_tab[i].type;

  for(i=0;solver_tab[i].name;i++)
    if(!strcmp(solver_tab[i].name,solver))
      if(DOIHAVE(solver_tab[i].code) == 0) {
	fprintf(stderr,"This version of MINOPT was not compiled with solver %s\n",solver);
	return -1;
      }
      else if(CANITDO(solver_tab[i].cap,type) == 0) {
	fprintf(stderr,"Solver %s cannot solve %s problem\n",solver,type_string);
	return -1;
      }
      else
	switch(type) {
	case LPTYPE:
	  SET_LPSOLV(solver_tab[i].code);
	  return 0;
        case NLPTYPE:
	  SET_NLPSOLV(solver_tab[i].code);
	  return 0;
        case MILPTYPE:
	  SET_MILPSOLV(solver_tab[i].code);
          return 0;
	case OCPTYPE:
	  SET_OCPSOLV(solver_tab[i].code);
	  return 0;
        case MINLPTYPE:
	  SET_MINLPSOLV(solver_tab[i].code);
          return 0;
	case DYNTYPE:
	  SET_DYNSOLV(solver_tab[i].code);
	  return 0;
	default:
	  fprintf(stderr,"Something is wrong.\n");
	  exit(0);
        }
  fprintf(stderr,"Solver %s is not a valid solver name\n",solver);
  return -1;
}

int get_solver(int type)
{
  switch(type) {
  case LPTYPE:
    return GET_LPSOLV();
  case NLPTYPE:
    return GET_NLPSOLV();
  case MILPTYPE:
    return GET_MILPSOLV();
  case OCPTYPE:
    return GET_OCPSOLV();
  case MINLPTYPE:
    return GET_MINLPSOLV();
  case DYNTYPE:
    return GET_DYNSOLV();
  default:
    fprintf(stderr,"Something is wrong.\n");
    exit(0);
  }
  return 0;
}

char *get_solver_string(int solver)
{
  int i;
  char *solver_string;
  for(i=0;solver_tab[i].name;i++)
    if(solver == solver_tab[i].code) {
      solver_string = (char *)malloc((strlen(solver_tab[i].name)+1)*sizeof(char));
      strcpy(solver_string,solver_tab[i].name);
      break;
    }
  return solver_string;
}

/* functions to get the solver for the different problem types */
/* based on the problem to be solved and its type (LP, NLP, etc.)
   these funtions return the funtion pointed to the appropriate function */
prifuntype get_prifun(int problem, int type, char *sname)
{
  int i;
  long int solver;
  
  switch(type) {
  case LPTYPE:    solver = GET_LPSOLV(); break;
  case NLPTYPE:   solver = GET_NLPSOLV(); break;
  case MILPTYPE:  solver = GET_MILPSOLV(); break;
  case OCPTYPE:   solver = GET_OCPSOLV(); break;
  default: 
    fprintf(stderr,"Invalid problem type.\n");
    exit(0);
  }
  
  for(i=0;solver_tab[i].name;i++)
    if(solver_tab[i].code == solver) {
      if(sname) strcpy(sname,solver_tab[i].name);
      switch(problem) {
      case PRIMAL: return solver_tab[i].prifun;
      case IPRIMAL: return solver_tab[i].iprifun;
      }
    }
  return NULL;  
}

masfuntype get_masfun(int problem, int type, char *sname)
{
  int i;
  long int solver;
  
  switch(type) {
  case LPTYPE:    solver = GET_LPSOLV(); break;
  case NLPTYPE:   solver = GET_NLPSOLV(); break;
  case MILPTYPE:  solver = GET_MILPSOLV(); break;
  default:
    fprintf(stderr,"Invalid problem type.\n");
    exit(0);
  }
  for(i=0;solver_tab[i].name;i++)
    if(solver_tab[i].code == solver) {
      if(sname) strcpy(sname,solver_tab[i].name);
      switch(problem) {
      case GBDMASTER: return solver_tab[i].gbdmasfun;
      case OAERMASTER: return solver_tab[i].oaermasfun;
      case OAERAPMASTER: return solver_tab[i].oaerapmasfun;
      case DSUB1: return  solver_tab[i].dsub1fun;
      case DSUB2: return  solver_tab[i].dsub2fun;
      }
    }
  return NULL;  
}

dmsfuntype get_dmsfun(int problem, int type, char *sname)
{
  int i;
  long int solver;
  
  switch(type) {
  case LPTYPE:    solver = GET_LPSOLV(); break;
  case NLPTYPE:   solver = GET_NLPSOLV(); break;
  case MILPTYPE:  solver = GET_MILPSOLV(); break;
  case MINLPTYPE: solver = GET_MINLPSOLV(); break;
  default:
    fprintf(stderr,"Invalid problem type.\n");
    exit(0);
  }
  for(i=0;solver_tab[i].name;i++)
    if(solver_tab[i].code == solver) {
      if(sname) strcpy(sname,solver_tab[i].name);
      switch(problem) {
      case DMASTER: return  solver_tab[i].dmsfun;
      }
    }
  return NULL;  
}

prifuntype get_rlxfun(int problem, int type, char *sname)
{
  int i;
  long int solver;
  
  switch(type) {
  case LPTYPE:    solver = GET_LPSOLV(); break;
  case NLPTYPE:   solver = GET_NLPSOLV(); break;
  case OCPTYPE:   solver = GET_OCPSOLV(); break;
  default:
    fprintf(stderr,"Invalid problem type.\n");
    exit(0);
  }
  for(i=0;solver_tab[i].name;i++)
    if(solver_tab[i].code == solver) {
      if(sname) strcpy(sname,solver_tab[i].name);
      switch(problem) {
      case RELAX: return  solver_tab[i].rlxfun;
      }
    }
  return NULL;  
}

minfuntype get_minfun(int problem, int type, char *sname)
{
  int i;
  long int solver;
  
  switch(type) {
  case MINLPTYPE: solver = GET_MINLPSOLV(); break;
  default:
    fprintf(stderr,"Invalid problem type.\n");
    exit(0);
  }
  for(i=0;solver_tab[i].name;i++)
    if(solver_tab[i].code == solver) {
      if(sname) strcpy(sname,solver_tab[i].name);
      switch(problem) {
      case MINLP: return  solver_tab[i].minfun;
      }
    }
  return NULL;  
}

dynfuntype get_dynfun(int problem, int type, char *sname)
{
  int i;
  long int solver;
  
  switch(type) {
  case DYNTYPE: solver = GET_DYNSOLV(); break;
  default:
    fprintf(stderr,"Invalid problem type.\n");
    exit(0);
  }
  for(i=0;solver_tab[i].name;i++)
    if(solver_tab[i].code == solver) {
      if(sname) strcpy(sname,solver_tab[i].name);
      switch(problem) {
      case DYNAMIC:  return solver_tab[i].dynfun;
      }
    }
  return NULL;  
}

adjfuntype get_adjfun(int problem, int type, char *sname)
{
  int i;
  long int solver;
  
  switch(type) {
  case DYNTYPE: solver = GET_DYNSOLV(); break;
  default:
    fprintf(stderr,"Invalid problem type.\n");
    exit(0);
  }
  for(i=0;solver_tab[i].name;i++)
    if(solver_tab[i].code == solver) {
      if(sname) strcpy(sname,solver_tab[i].name);
      switch(problem) {
      case ADJOINT:  return solver_tab[i].adjfun;
      }
    }
  return NULL;  
}

/* These functions are for the distribution */
/* If MINOS is used and not SNOPT or vice-versa,
   dummy functions must be provided */

#if defined(MINOS) && !defined(SNOPT)
void snopt_solve(void)
{
}
void snopt_set_option(void)
{
}
#endif

#if defined(SNOPT) && !defined(MINOS)
void minos_solve(void)
{
}
void minos_set_option(void)
{
}
#endif

#ifndef CPLEX
void close_cplex_env(void)
{
}
#endif
